MODULE isat_module

  ! global type definitions
  TYPE :: isat_record
     ! isat records
     ! accessed    = number of times the record is accessed
     ! start       = vector of independent variables
     !               "start" determines "finish"
     ! finish      = vector of dependent variables
     ! sensitivity = sensitivity matrix A(i,j) = d(finish)_(i) / d(start)_(j)
     ! eoa         = ellipsoid of accuracy around "start"
     ! stem        = pointer back to the binary tree stem
     INTEGER :: accessed
     DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: start
     DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: finish
     DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: sensitivity
     DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: eoa
     TYPE (isat_btree), POINTER :: stem
  END TYPE isat_record

  TYPE :: isat_btree
     ! binary tree
     ! the cutting plane is defined by  v' * y0 = a
     ! v      = transpose of the vector defining the cutting plane
     ! a      = scalar as part of the definition of the cutting plane
     ! lt     = pointer to the next node if v' * y0 < a
     ! gt     = pointer to the next node if v' * y0 >= a
     ! leaf   = pointer to the index of leaves in "isat_records"
     !          this pointer is nullified if it points to another branch
     DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: v
     DOUBLE PRECISION :: a
     TYPE (isat_btree), POINTER :: lt
     TYPE (isat_btree), POINTER :: gt
     TYPE (isat_record), POINTER :: leaf
  END TYPE isat_btree

  TYPE :: isat_btree_ptrs
     ! this is a work-around f90's lack of support for pointer arrays
     TYPE (isat_btree), POINTER :: ptr
  END TYPE isat_btree_ptrs

  TYPE :: isat_record_ptrs
     ! this is a work-around f90's lack of support for pointer arrays
     TYPE (isat_record), POINTER :: ptr
  END TYPE isat_record_ptrs

  TYPE :: isat_finish_constraints
     ! constraints on the finish variables
     LOGICAL :: active
     DOUBLE PRECISION :: lower
     DOUBLE PRECISION :: upper
  END TYPE isat_finish_constraints

  TYPE :: isat_raw_data
     ! used if isat_scalc = 1
     DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: start
     DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: finish
     TYPE (isat_raw_data), POINTER :: next
  END TYPE isat_raw_data
     
  ! global variable declarations

  ! pointers to the root of each binary tree
  ! dimensioned in isat_initialize
  TYPE (isat_btree_ptrs), DIMENSION (:), ALLOCATABLE :: isat_btree_roots

  ! pointer to the first of the raw data storage
  TYPE (isat_raw_data), POINTER :: isat_raw_head

  ! pointer to the last of the raw data storage (where to add next)
  TYPE (isat_raw_data), POINTER :: isat_raw_tail
   
  ! dimensioned in isat_initialize
  TYPE (isat_finish_constraints), DIMENSION (:), ALLOCATABLE :: isat_fcons

  ! default number of binary trees that are created when isat is initialized 
  INTEGER :: isat_binary_trees = 8

  ! keep track of which binary tree will receive the next new leaf (or record)
  INTEGER :: isat_btree_add = 1

  ! counter for the number of isat records
  INTEGER :: isat_records = 0

  ! has isat been initialized?
  LOGICAL :: isat_initialized = .FALSE.

  ! growth checking
  ! for method #1 this finds isat records that are inside the grown eoa
  !   It then contracts the eoa if it finds records outside the finish 
  !   error tolerance.
  ! for methods #2 and #3 this "looks both ways" and rejects the growth if
  !   the opposite side is outside the finish error tolerance.
  LOGICAL :: isat_do_growth_check = .TRUE.

  ! raw data record limit to avoid filling up memory
  INTEGER :: isat_raw_limit = 100000000
  INTEGER :: isat_raw_count = 0

  ! print level
  INTEGER :: isat_print_level = 0

  ! method of sensitivity calculation
  ! method #1 (default): estimate sensitivities from historical data
  !   this method tries to estimate the sensitivities from data that
  !     has already been computed
  !   prior calculations can be loaded to help obtain the sensitivities
  ! method #2: finite difference approximation
  !   this method requires an interface to the calculation
  !     finish = f(start) so that ISAT can provide various vectors
  !     of start to compute vectors of finish
  ! method #3: sensitivity calculation provided with isat_compute_finish
  !   this method requires an interface to the calculation
  !     finish = f(start) and to sensitivity = d(finish)/d(start)
  !   the sensitivity is calculated in isat_compute_finish subroutine
  INTEGER :: isat_scalc = 1

! isat subroutines
CONTAINS

  SUBROUTINE isat (ns,start,nf,finish,la,a,atol,opt)    
    ! the main ISAT routine
    ! supporting functions and global variables begin with "isat_"
    IMPLICIT NONE
    ! description of the variables
    INTEGER, INTENT (IN) :: ns, nf, la
    ! ns = number of predictor variables called "start"
    ! nf = number of predicted variables called "finish"
    ! la = leading dimension of the sensitivity matrix "a"
    INTEGER, INTENT (IN OUT) :: opt
    ! opt = isat option
    !    if opt = any integer expect 4 or 5 on input then isat will
    !       decide the scenario of retrieval, growth, or addition based
    !       on the error control
    !       on output opt will indicate the isat scenario selected
    !       if opt = 1 on output then a retrieval was performed
    !       if opt = 2 on output then a growth was performed
    !       if opt = 3 on output then an addition was performed
    !       if opt = -1 on output then an error occured
    !       if opt = -3 on output then an addition was requested by
    !         the automatic error control but did not happen.  Possible
    !         reasons include:
    !         o The database exceeds the current memory limit
    !         o An accurate sensitivity could not be computed
    !    the user also has the ability to force an isat operation
    !    on output opt will be unchanged for opt = 4
    !    for opt = 5, the output opt will be changed to 0
    !    if opt = 4 on input then a retrieval is forced
    !       in this case the states will not be bounded by "atol" and
    !         the sensitivities will be approximated by a stored record
    !    if opt = 5 on input then the isat database will be 
    !       initialized and return an exact "finish" and sensitivity if
    !       isat_scalc = 2 or 3
    !    if opt = 6 on input then the isat database will just read
    !       in raw data (sets of start-finish).  It is important to
    !       initialize the isat database with opt = 5 before calling
    !       this option.
    DOUBLE PRECISION, INTENT (IN) :: start(*), atol(*)
    DOUBLE PRECISION, INTENT (INOUT) :: finish(*)
    DOUBLE PRECISION, INTENT (OUT) :: a(la,*)
    
    ! local variables
    INTEGER :: i,j,k,error
    ! i,j,k = do loop variables
    LOGICAL :: within_start_eoa(isat_binary_trees)
    ! within_start_eoa = indicates if the query point is within the
    !   ellipsoid of accuracy
    LOGICAL :: within_finish_atol(isat_binary_trees)
    ! within_finish_atol = indicates if the final estimate is within atol 
    TYPE (isat_record_ptrs), DIMENSION(:), ALLOCATABLE :: leaves
    ! pointers to binary tree leaves (isat records)
    TYPE (isat_record), POINTER :: leaf
    TYPE (isat_record), POINTER :: leaf_added
    ! pointer to one leaf in the binary tree
    DOUBLE PRECISION :: atol_sq
    DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: eps_sq
    ! atol_sq = atol' * atol
    ! eps_sq = Delta_start' * eoa * Delta_start

    ! work variables to store intemediate results
    LOGICAL :: lwork
    INTEGER :: iwork
    DOUBLE PRECISION :: work_1
    DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: work_ns_1, work_ns_2
    DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: work_nf_1, work_nf_2
    DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: work_ns_ns
    DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: work_nf_ns

    INTRINSIC ABS, ASSOCIATED

    ALLOCATE (work_ns_1(ns),work_ns_2(ns))
    ALLOCATE (work_nf_1(nf),work_nf_2(nf))
    ALLOCATE (work_ns_ns(ns,ns))
    ALLOCATE (work_nf_ns(nf,ns))
    ALLOCATE (leaves(isat_binary_trees))
    ALLOCATE (eps_sq(isat_binary_trees))
    ! work arrays of size ns, nf, (ns x ns), etc.

    ! begin computational routine

    ! compute atol_sq = atol'*atol
    atol_sq = 0.0D0
    DO i = 1,nf
       atol_sq = atol(i)**2 + atol_sq
    ENDDO

    ! isat database initialization
    IF ((opt == 5).OR.(isat_initialized.EQV..FALSE.)) THEN
       CALL isat_initialize(nf)
    ENDIF

    IF (opt == 6) THEN
       IF (isat_raw_count >= isat_raw_limit) THEN
          PRINT *, 'WARNING: over isat_raw_limit - raw data not added'
          PRINT *, 'Current value of isat_raw_limit: ', isat_raw_limit
       ELSE
          CALL isat_add_raw(ns,start,nf,finish)
       ENDIF
       GO TO 100
    ENDIF

    ! search the binary trees for near records
    CALL isat_bt_search(ns,start,leaves)

    ! determine if start is within the eoa
    ! eps_sq = (start - stored_start)' * eoa * (start - stored_start)
    DO i = 1,isat_binary_trees
       IF (ASSOCIATED (leaves(i)%ptr)) THEN
          leaf => leaves(i)%ptr
          DO j = 1,ns
             work_ns_1(j) = start(j) - leaf%start(j)
             DO k = 1,ns
                work_ns_ns(j,k) = leaf%eoa(j,k)
             ENDDO
          ENDDO
          ! work_ns_2 = work_ns_ns * work_ns_1
          CALL DGEMV ('N', ns, ns, 1.0D0, work_ns_ns, ns, work_ns_1, 1, &
               0.0D0, work_ns_2, 1)
          ! eps_sq = work_ns_1' * work_ns_2
          eps_sq(i) = 0.0D0
          DO j = 1,ns
             eps_sq(i) = eps_sq(i) + (work_ns_1(j) * work_ns_2(j))
          ENDDO
          IF (eps_sq(i) <= atol_sq) THEN
             within_start_eoa(i) = .TRUE.
          ELSE
             within_start_eoa(i) = .FALSE.
          ENDIF
       ELSE
          within_start_eoa(i) = .FALSE.
       ENDIF
    ENDDO

    ! test if at least one within_start_eoa is .TRUE. and select the one
    !   with the lowest eqs_sq
    lwork = .FALSE.
    iwork = 0
    work_1 = atol_sq + 1.0D0
    DO i = 1,isat_binary_trees
       IF (within_start_eoa(i).EQV..TRUE.) THEN
          lwork = .TRUE.
          IF (eps_sq(i) < work_1) THEN
             iwork = i
             work_1 = eps_sq(i)
          ENDIF
       ENDIF
    ENDDO

    ! In case none are selected, select the best one for 
    !   opt = 4 (forced retrieval)
    IF ((iwork.EQ.0.).AND.(opt.EQ.4)) THEN
       work_1 = 1.0D30
       DO i = 1,isat_binary_trees
          IF (ASSOCIATED (leaves(i)%ptr)) THEN
             IF (eps_sq(i) < work_1) THEN
                iwork = i
                work_1 = eps_sq(i)
             ENDIF
          ENDIF
       ENDDO
       IF (iwork == 0) THEN
          PRINT *, 'A retrieval cannot be forced since there are no records'
          PRINT *, 'Changing opt to 0 (automatic error control)'
          opt = 0
       ENDIF
    ENDIF

    ! scenario #1: retrieval
    IF ((lwork.EQV..TRUE.).OR.(opt.EQ.4)) THEN
       leaf => leaves(iwork)%ptr

       call isat_retrieve (ns,start,nf,finish,leaf)

       ! add one to 'accessed'
       leaf%accessed = leaf%accessed + 1

       ! load the stored sensitivity matrix
       DO i = 1,nf
          DO j = 1,ns
             a(i,j) = leaf%sensitivity(i,j)
          ENDDO
       ENDDO

       ! indicate that isat performed a retrieval
       IF (opt == 4) THEN
          ! leave opt unchanged
       ELSE
          opt = 1
       ENDIF

       GO TO 100
    ENDIF

    ! the code has reaches this point when none of the records are within
    !   the start eoa - at this point growths are attempted
    ! determine if any of the records are within the finish tolerance (atol)
    ! lwork indicates if a sensitivity should be computed
    lwork = .FALSE.
    CALL isat_compute_finish(ns,start,nf,finish,la,a,lwork,error)

    IF (isat_scalc == 1) THEN
       IF (isat_raw_count >= isat_raw_limit) THEN
          PRINT *, 'WARNING: over isat_raw_limit - raw data not added'
          PRINT *, 'Current value of isat_raw_limit: ', isat_raw_limit
       ELSE
          CALL isat_add_raw(ns,start,nf,finish)
       ENDIF
    ENDIF

    ! initialize within_finish_atol
    DO i = 1,isat_binary_trees
       within_finish_atol(i) = .FALSE.
    ENDDO

    ! determine if the difference is within the finish tolerance
    DO i = 1,isat_binary_trees
       IF (ASSOCIATED (leaves(i)%ptr)) THEN
          ! estimate "finish"
          leaf => leaves(i)%ptr
          call isat_retrieve (ns,start,nf,work_nf_1,leaf)

          ! compute difference between the estimated and actual finish
          DO j = 1,nf
             work_nf_2(j) = ABS(finish(j) - work_nf_1(j))
          ENDDO

          within_finish_atol(i) = .TRUE.
          j = 0
          DO WHILE ((j <= nf - 1) .AND. (within_finish_atol(i).EQV..TRUE.))
             j = j + 1
             IF (work_nf_2(j) >= atol(j)) THEN
                within_finish_atol(i) = .FALSE.
             ENDIF
          ENDDO

          IF (within_finish_atol(i).EQV..TRUE.) THEN

             ! attempt growth of the eoa
             CALL isat_growth (ns,start,atol_sq,leaf)

             IF (isat_do_growth_check) THEN
                CALL isat_growth_check(ns,nf,atol,atol_sq,leaf)
             ENDIF

          ENDIF
       ENDIF
    ENDDO

    ! scenario #2: growth
    DO i = 1,isat_binary_trees  
       IF (within_finish_atol(i).EQV..TRUE.) THEN
          leaf => leaves(i)%ptr
          ! provide an estimate of the sensitivities
          DO j = 1,nf
             DO k = 1,ns
                a(j,k) = leaf%sensitivity(j,k)
             ENDDO
          ENDDO
          ! exact finish is returned
          opt = 2
          ! return the sensitivity of the first record with
          !   within_finish_atol(i).EQV..TRUE.
          ! a better estimate may be obtained through other methods
          !   such as taking an average of the sensitivities
          GO TO 100
       ENDIF
    ENDDO

    ! scenario #3: addition
    SELECT CASE (isat_scalc)
    CASE (1)
       CALL isat_est_sens(ns,start,nf,finish,la,a,atol,error)
    CASE (2)
       PRINT *, 'isat_scalc=2: functionality not yet added'
       error = 1
       !CALL isat_comp_sens(ns,start,nf,finish,la,a,error)
    CASE (3)
       ! lwork = .TRUE. tells isat_compute_finish to calculate the 
       !   sensitivity matrix
       lwork = .TRUE.
       CALL isat_compute_finish(ns,start,nf,finish,la,a,lwork,error)
    CASE DEFAULT
       PRINT *, 'Invalid current value of isat_scalc = ', isat_scalc
       PRINT *, 'Changing to isat_scalc = 1 (default)'
       isat_scalc = 1
    END SELECT

    IF (error.EQ.0) THEN
       ! add a leaf and branch to the binary tree
       leaf => leaves(isat_btree_add)%ptr
       CALL isat_add(ns,start,nf,finish,la,a,atol_sq,leaf,leaf_added)
       ! exact final states and exact sensitivities are returned with option 3
       IF (isat_do_growth_check) THEN
          CALL isat_growth_check(ns,nf,atol,atol_sq,leaf_added)
       ENDIF
       opt = 3
    ELSE
       opt = -3
    ENDIF

100 CONTINUE

    DEALLOCATE (work_ns_1,work_ns_2)
    DEALLOCATE (work_nf_1,work_nf_2)
    DEALLOCATE (work_ns_ns)
    DEALLOCATE (work_nf_ns)
    DEALLOCATE (leaves)
    DEALLOCATE (eps_sq)

    RETURN
  END SUBROUTINE isat


  ! initialize the isat database
  SUBROUTINE isat_initialize(nf)

    IMPLICIT NONE

    INTEGER :: nf

    ! local variables
    INTEGER :: i

    IF (isat_initialized.EQV..TRUE.) CALL isat_unload
 
    ! create isat roots (allocate array of pointers)
    ALLOCATE (isat_btree_roots(isat_binary_trees))

    ! nullify pointers to the data
    DO i = 1,isat_binary_trees
       NULLIFY (isat_btree_roots(i)%ptr)
    ENDDO

    ! allocate finish constraint storage
    ALLOCATE (isat_fcons(nf))
    DO i = 1,nf
       isat_fcons(i)%active = .FALSE.
    ENDDO

    ! indicate that tree 1 should be added to first
    isat_btree_add = 1

    ! indicate that the ISAT database has been initialized
    isat_initialized = .TRUE.

    ! set up raw data storage in case isat_scalc = 1 is selected 
    ALLOCATE (isat_raw_head)
    isat_raw_tail => isat_raw_head
    NULLIFY(isat_raw_tail%next)

    RETURN
  END SUBROUTINE isat_initialize


  ! unload isat memory
  SUBROUTINE isat_unload
    ! use a recursive function to unload memory
    ! look at Redwine's advice about depth of recursiveness

    ! this code hasn't been completed

    isat_initialized = .FALSE.

    RETURN
  END SUBROUTINE isat_unload


  ! isat binary tree search
  SUBROUTINE isat_bt_search(ns,start,leaves)
    ! Binary Tree Search
    ! Description:  This subroutine searches the binary tree for a 
    !   close record.  The tree consists of branches and leaves at
    !   the ends of the branches. 
    
    ! Method:  Cutting planes are defined by  v' * start = a.  Continually
    !   divide the database until one record is left.

    ! This subroutine returns record pointers from multiple binary trees

    IMPLICIT NONE

    INTEGER, INTENT (IN) :: ns
    DOUBLE PRECISION, INTENT (IN) :: start(*)

    TYPE (isat_record_ptrs) :: leaves(*)

    ! local variables
    INTEGER          :: i,j
    ! i,j = index variables
    DOUBLE PRECISION :: work
    TYPE (isat_btree), POINTER :: btree_ptr

    INTRINSIC ASSOCIATED

    DO i = 1,isat_binary_trees
       IF (ASSOCIATED (isat_btree_roots(i)%ptr)) THEN
          ! start with the root
          btree_ptr => isat_btree_roots(i)%ptr

          ! as long as the branch is not a leaf, then divide again
          DO WHILE (.NOT. ASSOCIATED (btree_ptr%leaf))
             ! compute v' * start
             work = 0.0D0
             DO j = 1,ns
                work = work + (btree_ptr%v(j) * start(j))
             ENDDO
             ! compare with "a" 
             IF (work < btree_ptr%a) THEN
                btree_ptr => btree_ptr%lt
             ELSE
                btree_ptr => btree_ptr%gt
             ENDIF
          END DO
          leaves(i)%ptr => btree_ptr%leaf
       ELSE
          NULLIFY (leaves(i)%ptr)
       ENDIF
    ENDDO

    RETURN
  END SUBROUTINE isat_bt_search


  ! isat retrieval
  SUBROUTINE isat_retrieve (ns,start,nf,finish,leaf) 
    INTEGER, INTENT (IN) :: ns, nf
    DOUBLE PRECISION, INTENT (IN) :: start(*)
    DOUBLE PRECISION, INTENT (OUT) :: finish(*)
    TYPE (isat_record), POINTER :: leaf

    ! work variables
    DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: work_ns
    DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: work_nf
    DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: work_nf_ns

    ! allocate work space
    ALLOCATE (work_ns(ns),work_nf(nf),work_nf_ns(nf,ns))

    ! estimate "finish"
    ! finish = stored_finish + (stored_sensitivity * (start - stored_start))

    ! load the stored sensitivity matrix
    DO i = 1,nf
       DO j = 1,ns
          work_nf_ns(i,j) = leaf%sensitivity(i,j)
       ENDDO
    ENDDO
    ! load the stored finish
    DO i = 1,nf
       work_nf(i) = leaf%finish(i)
    ENDDO
    ! compute the difference (start - stored_start)
    DO i = 1,ns
       work_ns(i) = start(i) - leaf%start(i)
    ENDDO

    ! work_nf = work_nf_ns * work_ns + work_nf
    CALL DGEMV ('N', nf, ns, 1.0D0, work_nf_ns, nf, work_ns, 1, & 
         1.0D0, work_nf, 1)

    DO i = 1,nf
       IF (isat_fcons(i)%active.EQV..TRUE.) THEN
          IF (isat_fcons(i)%upper < isat_fcons(i)%lower) THEN
             PRINT *, 'Error: upper bound is less than lower bound for constraint ', i
          ENDIF
          IF (work_nf(i) < isat_fcons(i)%lower) THEN
             finish(i) = isat_fcons(i)%lower
          ENDIF
          IF (work_nf(i) > isat_fcons(i)%upper) THEN
             finish(i) = isat_fcons(i)%upper
          ENDIF
       ELSE
          finish(i) = work_nf(i)
       ENDIF
    ENDDO

    DEALLOCATE (work_ns,work_nf,work_nf_ns)

    RETURN
  END SUBROUTINE isat_retrieve


  ! isat growth
  SUBROUTINE isat_growth (ns,start,atol_sq,leaf)
    ! grow binary tree leaf

    IMPLICIT NONE

    INTEGER, INTENT (IN) :: ns
    DOUBLE PRECISION, INTENT (IN) :: start(*)
    DOUBLE PRECISION, INTENT (IN) :: atol_sq
    TYPE (isat_record), POINTER :: leaf
    
    ! local variables
    INTEGER :: i,j, info, expand, iwork_dsyev
    ! i,j = index variables
    ! info = message codes from lapack subroutines
    ! expand = the principle direction to expand
    ! iwork_dsyev = size of work_dsyev
    DOUBLE PRECISION :: smallest_sigma, t1_dstart_mag
    DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: dstart, sigma
    DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: work_dsyev, work_ns
    DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: t1_dstart
    DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: t1_dstart_dir
    DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: eoa, m2
    DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: q, r
    DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: t1, t2, onb
    ! dstart = (start - start_stored)
    ! eoa = ellipsoid of accuracy
    ! q = Schur decomposition eigenvectors
    ! r = r in the qr factorization form
    ! sigma = vector of eigenvalues
    ! work_dsyev = work vector for lapack routine dsyev
    !   (dimensioned 2x required)
    ! work_ns = work vector of length ns
    ! t1 = rotation matrix 1
    ! t2 = rotation matrix 2
    ! t1_dstart = dstart transformed to the unit hyper-sphere coordinates
    ! t1_dstart_mag = length of t1_dstart, measured by the 2-norm
    ! t1_dstart_dir = direction of t1_dstart as a unit vector
    ! onb = matrix used to generate an orthonormal basis to t1_dstart
    ! smallest_sigma = smallest eigenvalue of onb
    ! m2 = unit hyper-spheroid with one expanded axis length

    ! intrinsic functions
    INTRINSIC ABS, SQRT

    ! size of vector work_dsyev
    iwork_dsyev = 6*ns

    ALLOCATE (dstart(ns))
    ALLOCATE (eoa(ns,ns))
    ALLOCATE (q(ns,ns), r(ns,ns), sigma(ns))
    ALLOCATE (work_dsyev(iwork_dsyev), work_ns(ns))
    ALLOCATE (t1(ns,ns), t2(ns,ns))
    ALLOCATE (t1_dstart(ns))
    ALLOCATE (t1_dstart_dir(ns), onb(ns,ns))
    ALLOCATE (m2(ns,ns))

    ! compute dstart
    DO i = 1,ns
       dstart(i) = start(i) - leaf%start(i)
    ENDDO

    ! retrieve original eoa
    DO i = 1,ns
       DO j = 1,ns
          eoa(i,j) = leaf%eoa(i,j)
       ENDDO
    ENDDO

    ! perform a Schur decomposition of the eoa
    ! eoa = q * T * q'
    ! the eigenvalues are the diagonal of T
    ! for a symmetric matrix, T is strictly diagonal
    ! an eigenvalue decomposition is equivalent to a schur decomposition
    !   when the starting matrix is symmetric - it is much easier to
    !   use the lapack routine DGESV rather than DGEES
    ! compute an orthonormal basis of eoa
    CALL DSYEV('V','U',ns,eoa,ns,sigma,work_dsyev,iwork_dsyev,info)

    DO i = 1,ns
       DO j = 1,ns
          q(i,j) = eoa(i,j)
       ENDDO
    ENDDO
    ! calculate similarity transform to map hyper-ellipsoid to a 
    !   unit hyper-sphere
    ! t1 = diag(sqrt(sigma/atol_sq)) * q'
    ! 1st step (sqrt(sigma/atol_sq))
    DO i = 1,ns
       work_ns(i) = SQRT(ABS(sigma(i)/atol_sq))
    ENDDO
    ! 2nd step (diag(work_ns) * q')
    DO i = 1,ns
       DO j = 1,ns
          t1(i,j) = work_ns(i) * q(j,i)
       ENDDO
    ENDDO
    
    ! transform dstart into the unit hyper-sphere coordinates
    ! t1_dstart = t1 * dstart
    CALL DGEMV ('N',ns,ns,1.0D0,t1,ns,dstart,1,0.0D0,t1_dstart,1)
    
    ! determine the magnitude and direction of t1_dstart
    t1_dstart_mag = 0.0D0
    DO i = 1,ns
       t1_dstart_mag = t1_dstart_mag + t1_dstart(i)**2
    ENDDO
    t1_dstart_mag = SQRT(t1_dstart_mag)
    DO i = 1,ns
       t1_dstart_dir(i) = t1_dstart(i) / t1_dstart_mag
    ENDDO
    ! generate an orthonormal basis with t1_dstart as one of the
    !   column vectors (it is the one that corresponds to the
    !   zero eigenvalue)
    ! onb = eye(n) - t1_dstart_dir * t1_dstart_dir'
    DO i = 1,ns
       DO j = 1,ns
          onb(i,j) = - t1_dstart_dir(i) * t1_dstart_dir(j)
       ENDDO
       onb(i,i) = onb(i,i) + 1.0D0
    ENDDO
    
    ! compute an orthonormal basis to t1_dstart
    CALL DSYEV('V','U',ns,onb,ns,sigma,work_dsyev,iwork_dsyev,info)
    
    ! rotate the unit hyper-sphere to align with the external point
    ! t2 = inv(onb) = transpose(onb)
    DO i = 1,ns
       DO j = 1,ns
          t2(j,i) = onb(i,j)
       ENDDO
    ENDDO
    
    ! the axis to be expanded is the one that corresponds to the zero
    !   eigenvalue, while the other eigenvalues should be equal to one
    ! this routine just selects the eigenvalue position that is closest
    !   to zero
    
    smallest_sigma = 1.0D0
    ! bandaid for a bug:
    ! the code crashed when onb had a high condition number
    ! this resulted in no selection of variable expand and therefore 
    !   a problem a few lines down
    ! one solution is to not grow the eoa but just accept the extra
    !   cost of integration by not expanding the eoa
    expand = -1
    DO i = 1,ns
       sigma(i) = ABS(sigma(i))
       IF (sigma(i) < smallest_sigma) THEN
          expand = i
          smallest_sigma = sigma(i)
       ENDIF
    ENDDO
    
    IF (expand == -1) THEN
       !    don't expand the ellipsoid
       IF (isat_print_level>=1) THEN
          PRINT *, 'Orthonormal basis close to singular: EOA not expanded'
       ENDIF
       GO TO 110
    ENDIF
    
    ! minimum volume expansion of the unit hyper-sphere including dstart
    DO i = 1,ns
       DO j = 1,ns
          m2(i,j) = 0.0D0
       ENDDO
       m2(i,i) = 1.0D0
    ENDDO
    m2(expand,expand) = 1.0D0/(t1_dstart_mag**2)
    
    ! transform the ellipsoid back to the original coordinates
    ! eoa = atol_sq * (t1'*t2'*m2*t2*t1)
    ! 1st step (q = t2 * t1)
    CALL DGEMM ('N','N',ns,ns,ns,1.0D0,t2,ns,t1,ns,0.0D0,q,ns)
    ! 2nd step (r = m2 * q)
    CALL DGEMM ('N','N',ns,ns,ns,1.0D0,m2,ns,q,ns,0.0D0,r,ns)
    ! 3rd step (eoa = atol_sq * (q' * r))
    CALL DGEMM ('T','N',ns,ns,ns,atol_sq,q,ns,r,ns,0.0D0,eoa,ns)
    !

    ! store the expanded eoa
    DO i = 1,ns
       DO j = 1,ns
          leaf%eoa(i,j) = eoa(i,j)
       ENDDO
    ENDDO

110 CONTINUE

    DEALLOCATE (dstart)
    DEALLOCATE (eoa)
    DEALLOCATE (q, r, sigma)
    DEALLOCATE (work_dsyev, work_ns)
    DEALLOCATE (t1, t2)
    DEALLOCATE (t1_dstart)
    DEALLOCATE (t1_dstart_dir, onb)
    DEALLOCATE (m2)
    
    RETURN
  END SUBROUTINE isat_growth


  ! isat addition
  SUBROUTINE isat_add(ns,start,nf,finish,la,a,atol_sq,old_leaf,new_leaf)

    IMPLICIT NONE

    INTEGER, INTENT (IN) :: ns, nf, la
    DOUBLE PRECISION, INTENT (IN) :: atol_sq
    DOUBLE PRECISION, INTENT (IN) :: start(*)
    DOUBLE PRECISION, INTENT (IN) :: finish(*)
    DOUBLE PRECISION, INTENT (IN) :: a(la,*)
    TYPE (isat_record), POINTER :: old_leaf

    ! local variables
    INTEGER :: i,j,k
    INTEGER :: iwork_1, iwork_2
    DOUBLE PRECISION :: work_1
    DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: work_ns_1
    DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: work_ns_2
    DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: work_nf
    DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: work_svd
    DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: work_nf_ns   
    DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: work_nf_nf   
    DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: work_ns_ns
    TYPE (isat_btree), POINTER :: btree_ptr
    TYPE (isat_btree), POINTER :: btree_new_stem_1
    TYPE (isat_btree), POINTER :: btree_new_stem_2
    TYPE (isat_record), POINTER :: new_leaf

    ! intrinsic functions
    INTRINSIC ASSOCIATED, MAX, MIN, MOD

    ALLOCATE (work_ns_1(ns), work_ns_2(ns), work_nf(nf))
    ALLOCATE (work_nf_nf(nf,nf), work_nf_ns(nf,ns), work_ns_ns(ns,ns))

    ! compute the an initial estimate of the eoa based on a 
    !   zero-order approximation
    ! load the sensitivity matrix "a" into work_nf_ns
    DO i = 1,nf
       DO j = 1,ns
          work_nf_ns(i,j) = a(i,j)
       ENDDO
    ENDDO
    ! perform a singular value decomposition
    ! 5 times the required work space for svd
    iwork_1 = 5 * MAX(3*MIN(nf,ns)+MAX(nf,ns),5*MIN(nf,ns))
    ALLOCATE (work_svd(iwork_1))
    CALL DGESVD('N','A',nf,ns,work_nf_ns,nf,work_nf,work_nf_nf,nf, &
         work_ns_ns,ns,work_svd,iwork_1,iwork_2)
    DEALLOCATE (work_svd)
    ! ensure that none of the principal directions have an axis length
    !   that is too large
    DO i = 1,nf
       work_nf(i) = MAX(atol_sq/2.0D0,work_nf(i))
    ENDDO
    ! the new augmented sensitivity matrix becomes
    ! A_aug = U * diag(sigma) * VT
    ! the zero-order eoa estimate is
    ! eoa = A_aug' * A_aug
    ! this is the same as
    ! eoa = (VT' * diag(sigma) * U') * (U * diag(sigma) * VT)
    ! eoa = VT' * diag(sigma) * diag(sigma) * VT
    ! compute diag(sigma) * VT (multiply row i of VT by sigma(i))
    DO i = 1,nf
       DO j = 1,ns
          work_nf_ns(i,j) = work_nf(i) * work_ns_ns(i,j)
       ENDDO
    ENDDO
    ! compute zero-order eoa
    CALL DGEMM ('T', 'N', ns, ns, nf, 1.0D0, work_nf_ns, nf, work_nf_ns, &
         nf, 0.0D0, work_ns_ns, ns)
  
    ! special case for a root with no leaves (initialized root)
    IF (.NOT. ASSOCIATED (old_leaf)) THEN
       ! create nameless storage for a stem and point to it
       ALLOCATE (isat_btree_roots(isat_btree_add)%ptr)
       ! point to the root
       btree_ptr => isat_btree_roots(isat_btree_add)%ptr
       ! create nameless storage for a leaf and point to it
       ALLOCATE (btree_ptr%leaf)
       ! point to the new leaf
       new_leaf => btree_ptr%leaf
       ! point the new leaf back to its new stem
       new_leaf%stem => btree_ptr

       ! no branch information is required since the binary tree only
       !   consists of a root, stem, and one leaf

    ELSE
       ! compute information for the new branch
       ! define cutting plane (v and a)
       DO i = 1,ns
          work_ns_1(i) = start(i) - old_leaf%start(i)
          work_ns_2(i) = start(i) + old_leaf%start(i)
       ENDDO
       ! work_1 = 0.5 * (work_ns_1' * work_ns_2)
       work_1 = 0.0D0
       DO I = 1,ns
          work_1 = work_1 + (work_ns_1(i) * work_ns_2(i))
       ENDDO
       work_1 = work_1 * 0.5D0

       ! point to the old stem and transform it into a branch
       btree_ptr => old_leaf%stem
       ALLOCATE (btree_ptr%v(ns))
       DO i = 1,ns
          btree_ptr%v(i) = work_ns_1(i)
       ENDDO
       btree_ptr%a = work_1
       ! allocate storage for 2 new stems
       ALLOCATE (btree_ptr%lt)
       ALLOCATE (btree_ptr%gt)
       ! disassociate pointer to old_leaf
       NULLIFY (btree_ptr%leaf)

       ! point to the new location for the old leaf
       btree_new_stem_1 => btree_ptr%lt
       ! point to the old leaf
       btree_new_stem_1%leaf => old_leaf
       ! point the old leaf back at its new stem
       old_leaf%stem => btree_new_stem_1

       ! point to the location for the new leaf
       btree_new_stem_2 => btree_ptr%gt
       ! allocate storage for a new leaf and point to it
       ALLOCATE (btree_new_stem_2%leaf)
       new_leaf => btree_new_stem_2%leaf
       ! point the new leaf back at its new stem
       new_leaf%stem => btree_new_stem_2
    ENDIF

    ! fill in the new leaf information (except for the stem)
    new_leaf%accessed = 0
    ALLOCATE (new_leaf%start(ns))
    DO i = 1,ns
       new_leaf%start(i) = start(i)
    ENDDO
    ALLOCATE (new_leaf%finish(nf))
    DO i = 1,nf
       new_leaf%finish(i) = finish(i)
    ENDDO
    ALLOCATE (new_leaf%sensitivity(nf,ns))
    DO i = 1,nf
       DO j = 1,ns
          new_leaf%sensitivity(i,j) = a(i,j)
       ENDDO
    ENDDO
    ALLOCATE (new_leaf%eoa(ns,ns))
    DO i = 1,ns
       DO j = 1,ns
          new_leaf%eoa(i,j) = work_ns_ns(i,j)
       ENDDO
    ENDDO

    ! add one to the records count
    isat_records = isat_records + 1
    
    ! indicate which binary tree that will receive the next addition
    isat_btree_add = isat_btree_add + 1
    isat_btree_add = MOD(isat_btree_add-1,isat_binary_trees) + 1

    DEALLOCATE (work_ns_1, work_ns_2, work_nf)
    DEALLOCATE (work_nf_nf, work_nf_ns, work_ns_ns)

    RETURN
  END SUBROUTINE isat_add

  SUBROUTINE isat_add_raw (ns,start,nf,finish)
    ! this subroutine is used for the database management of the
    ! raw data when isat_scalc = 1.  Specifically, it adds one record
    ! to the end of the sequential list of raw data.

    INTEGER, INTENT (IN) :: ns, nf
    DOUBLE PRECISION, INTENT (IN) :: start(*)
    DOUBLE PRECISION, INTENT (IN) :: finish(*)

    INTEGER :: i
    TYPE (isat_raw_data), POINTER :: add

    ! global variable
    isat_raw_count = isat_raw_count + 1

    ! Point to the tail
    add => isat_raw_tail

    ALLOCATE (add%start(ns))
    ALLOCATE (add%finish(nf))

    DO i = 1,ns
       add%start(i) = start(i)
    ENDDO
    DO i = 1,nf
       add%finish(i) = finish(i)
    ENDDO
    ! allocate storage here for the next addition
    ! if a record has a nullified pointer to 'next' then
    !   it is the tail
    ALLOCATE (add%next)
    isat_raw_tail => add%next
    NULLIFY (isat_raw_tail%next)
    RETURN
  END SUBROUTINE isat_add_raw

  SUBROUTINE isat_est_sens (ns,start,nf,finish,la,a,atol,error)

    IMPLICIT NONE

    INTEGER, INTENT (IN) :: ns, nf, la
    INTEGER, INTENT (OUT) :: error
    INTEGER :: counter
    INTEGER :: i, j, ns_max, replace, remove, info
    INTEGER :: work_7_size
    INTEGER :: ld_work_9

    DOUBLE PRECISION, INTENT (IN) :: start(*)
    DOUBLE PRECISION, INTENT (IN) :: finish(*)
    DOUBLE PRECISION, INTENT (OUT) :: a(la,*)
    DOUBLE PRECISION, INTENT (IN) :: atol(*)
    DOUBLE PRECISION :: worst_1norm, worst_resid
    DOUBLE PRECISION :: test_1norm, temp_1norm
    DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: best_1norm
    DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: best_start
    DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: best_finish
    DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: work_1
    DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: work_2
    DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: work_3
    DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: work_4
    DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: work_5
    DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: work_6
    DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: work_7
    DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: work_8
    DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: work_9
    DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: work_10
    DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: work_11
    DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: work_12
    DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: work_13
    DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: work_14

    TYPE (isat_raw_data), POINTER :: data_ptr

    ! Alternating filter
    ! filter_type = .FALSE. to eliminate the highest residual
    ! filter_type = .TRUE. to eliminate the highest "d_start" 1-norm
    LOGICAL :: test
    LOGICAL :: prefilter = .FALSE.
    LOGICAL :: postfilter
    LOGICAL :: postfilter_type = .FALSE.

    INTRINSIC ABS, ASSOCIATED

    ! need a better test for prefilter
    ! eliminates data based on input vector closeness
    prefilter = .FALSE.

    ! max number of records to include in the regression
    ns_max = 4*ns

    ALLOCATE (best_1norm(ns_max))
    ALLOCATE (best_start(ns_max,ns),best_finish(ns_max,nf))

    counter = 0
    data_ptr => isat_raw_head

    ! Run through all of the data sequentially
    DO WHILE (ASSOCIATED(data_ptr%next))

       ! The section is a raw data pre-filter to collect the best records
       test = .TRUE.

       temp_1norm = 0.0D0
       DO i = 1,ns
          temp_1norm = temp_1norm + ABS(start(i)-data_ptr%start(i))
       ENDDO

       ! Exclude identical records
       IF (temp_1norm <= 1e-10) test = .FALSE.

       IF (prefilter.EQV..TRUE.) THEN
          ! I need a better test for this
          IF (temp_1norm > 10.d0 * atol(1) * ns) test = .FALSE.

          ! I could also include a test here to make sure the active
          !  sets are the same for the constrained manipulated variables

          DO i = 1,nf
             IF (ABS(finish(i)-data_ptr%finish(i)) > atol(i)) test = .FALSE.
          ENDDO
       ENDIF

       ! Record records that make it through the raw data pre-filter
       IF (test.EQV..TRUE.) THEN
          IF (counter < ns_max) THEN
             counter = counter + 1
             DO i = 1,ns
                best_start(counter,i) = data_ptr%start(i)
             ENDDO
             DO i = 1,nf
                best_finish(counter,i) = data_ptr%finish(i)
             ENDDO
             best_1norm(counter) = temp_1norm
          ELSE
             ! replace highest 1-norm with this one (if it is lower)
             ! identify the worst 1-norm record
             worst_1norm = 0.0D0
             DO i = 1,counter
                IF (best_1norm(i) > worst_1norm) THEN
                   worst_1norm = best_1norm(i)
                   replace = i
                ENDIF
             ENDDO
             
             IF (temp_1norm < worst_1norm) THEN
                DO i = 1,ns
                   best_start(replace,i) = data_ptr%start(i)
                ENDDO
                DO i = 1,nf
                   best_finish(replace,i) = data_ptr%finish(i)
                ENDDO
                best_1norm(replace) = temp_1norm
             ENDIF
          ENDIF
       ENDIF

       data_ptr => data_ptr%next
    ENDDO
          
11  CONTINUE

    ! When there are no records available return with error flag
    IF (counter == 0) THEN
       DEALLOCATE (best_1norm, best_start, best_finish)
       error = 1
       ! There are not enough good records
       IF (isat_print_level>=1) THEN
         PRINT *, 'Need ', 2*ns, ' raw records where ', counter, ' are available - can not add record'
       ENDIF
       RETURN
    ENDIF

    ! Dimension work arrays
    ld_work_9 = counter
    work_7_size = 10*counter
    ALLOCATE (work_1(ns,counter), work_2(nf,counter))
    ALLOCATE (work_3(ns,ns))
    ALLOCATE (work_5(ns,ns), work_6(ns,ns))
    ALLOCATE (work_8(ns,ns))
    ALLOCATE (work_9(counter,ns), work_10(nf,ns))
    ALLOCATE (work_11(nf,counter),work_12(nf,counter))

    ALLOCATE (work_4(ns),work_7(work_7_size))
    ALLOCATE (work_13(counter),work_14(counter))

    ! Subtract the origin to put the variables in deviation form
    ! and take a transpose
    DO i = 1,counter
       DO j = 1,ns
          work_1(j,i) = best_start(i,j) - start(j)
       ENDDO
       DO j = 1,nf
          work_2(j,i) = best_finish(i,j) - finish(j)
       ENDDO
    ENDDO
       
    ! Make a copy of best_1norm for filtering
    DO i = 1,counter
       work_14(i) = best_1norm(i)
    ENDDO

    DO WHILE (counter >= 2*ns)
       ! Compute statistical sensitivity
       ! i.e. K_est = u_obs * x_obs' * (x_obs * x_obs')^-1
       ! dimensions:
       ! u_obs = p x obs
       ! x_obs = n x obs
       ! K_est = (p x obs) * (obs x n) * (n x n) = p x n
       ! where u_obs is a matrix of "finish" observations
       ! where x_obs is a matrix of "start" observations
       
       ! Compute work_3 = x_obs * x_obs'
       CALL DGEMM ('N','T',ns,ns,counter,1.0D0,work_1,ns, &
            work_1,ns,0.0D0,work_3,ns)

       ! Compute (x_obs * x_obs')^-1
       ! Since x_obs * x_obs' may be rank deficient, a pseudo-inverse
       ! can be computed by inverting the singular values
       ! of a singular value decomposition
       ! M = U * S * V' = work_5 * work_4 * work_6
       CALL DGESVD('A','A',ns,ns,work_3,ns,work_4,work_5, &
            ns,work_6,ns,work_7,work_7_size,info)

       ! Compute S^-1 = work_4^-1
       DO i = 1,ns
          IF (work_4(i) > 1D-10) THEN
             work_4(i) = 1.0D0 / work_4(i)
          ELSE
             work_4(i) = 0.0D0
          ENDIF
       ENDDO
       ! M^-1 = V * (S^-1 * U') = V * (U * S^-1)'
       ! Compute U_new = (U * S^-1) first: work_5 = work_5 * work_4 
       DO i = 1,ns
          DO j = 1,ns
             work_5(i,j) = work_5(i,j) * work_4(j)
          ENDDO
       ENDDO

       ! Compute M^-1 = work_8 = work_6' * work_5'
       CALL DGEMM ('T','T',ns,ns,ns,1.0D0,work_6, &
            ns,work_5,ns,0.0D0,work_8,ns)

       ! Compute x_obs' * (M^-1): work_9 = work_1' * work_8
       CALL DGEMM ('T','N',counter,ns,ns,1.0D0,work_1,ns, &
            work_8,ns,0.0D0,work_9,ld_work_9)

       ! Compute work_10 = u * work_9
       CALL DGEMM ('N','N',nf,ns,counter,1.0D0,work_2,nf,work_9, &
            ld_work_9,0.0D0,work_10,nf)

       ! Compute the residuals (resid = pred - actual)
       ! Compute the predicted values: work_11 = work_10 * work_1
       CALL DGEMM ('N','N',nf,counter,ns,1.0D0,work_10,nf,work_1, &
            ns,0.0D0,work_11,nf)
       ! Compute the residuals work_12 = work_11 - work_2
       DO i = 1,nf
          DO j = 1,counter
             work_12(i,j) = work_11(i,j) - work_2(i,j)
          ENDDO
       ENDDO

       ! Determine if the residuals are within the final error tolerance
       postfilter = .FALSE.
       DO i = 1,counter
          work_13(i) = 0.0D0
          DO j = 1,nf
             IF (ABS(work_12(j,i)) > atol(j)) THEN
                ! Record the degree of violation for filtering
                work_13(i) = work_13(i) + ABS(work_12(j,i) - atol(j))
                postfilter = .TRUE.
             ENDIF
          ENDDO
       ENDDO

       IF (postfilter.EQV..TRUE.) THEN
          IF (postfilter_type.EQV..FALSE.) THEN
             ! identify highest residual
             worst_resid = 0.0D0
             DO i = 1,counter
                IF (worst_resid <= work_13(i)) THEN
                   remove = i
                   worst_resid = work_13(i)
                ENDIF
             ENDDO
          ELSE
             ! identify highest "d_start" 1-norm 
             worst_1norm = 0.0D0
             DO i = 1,counter
                IF (worst_1norm <= work_14(i)) THEN
                   remove = i
                   worst_resid = work_14(i)
                ENDIF
             ENDDO
          ENDIF

          ! Remove a record by shifting all records up one
          ! The last record will be dropped from the computation
          IF (remove < counter) THEN
             DO i = remove,counter - 1
                DO j = 1,ns
                   work_1(j,i) = work_1(j,i+1)
                ENDDO
                DO j = 1,nf
                   work_2(j,i) = work_2(j,i+1)
                ENDDO
                work_14(i) = work_14(i+1)
             ENDDO
          ENDIF

          ! alternate filter type
          postfilter_type = .NOT.postfilter_type

          ! decrease the number of available records counter
          counter = counter - 1
       ELSE
          EXIT
       ENDIF

    ENDDO

    IF (counter < (2*ns)) THEN
       error = 1
       ! There are not enough good records
       IF (isat_print_level>=1) THEN
          PRINT *, 'Need ', 2*ns, ' records where ', counter, ' are available - can not add record'
       ENDIF
    ELSE
       error = 0
       ! There are enough good records
       ! Transfer matrix work_10 to matrix a
       DO i = 1,nf
          DO j = 1,ns
             a(i,j) = work_10(i,j)
          ENDDO
       ENDDO

       IF (isat_print_level>=2) THEN
          PRINT *, ''
          PRINT *, 'SENSITIVITY (',nf,' x ',ns,')'
          DO i = 1,nf
             DO j = 1,ns
                WRITE(*,FMT='(1ES12.4)',ADVANCE='no') a(i,j)
             ENDDO
             WRITE(*,*)
          ENDDO
       ENDIF
    ENDIF

    DEALLOCATE (best_1norm, best_start, best_finish)
    DEALLOCATE (work_1,work_2,work_3)
    DEALLOCATE (work_5,work_6)
    DEALLOCATE (work_8,work_9)
    DEALLOCATE (work_10,work_11,work_12)

    DEALLOCATE (work_4,work_7,work_13,work_14)

    RETURN

  END SUBROUTINE isat_est_sens

  SUBROUTINE isat_growth_check(ns,nf,atol,atol_sq,leaf)
    IMPLICIT NONE

    INTEGER, INTENT (IN) :: ns,nf
    DOUBLE PRECISION, INTENT (IN) :: atol(*)
    DOUBLE PRECISION, INTENT (IN) :: atol_sq
    TYPE (isat_record), POINTER :: leaf

    INTEGER :: i,j

    DOUBLE PRECISION :: eps_sq
    DOUBLE PRECISION, ALLOCATABLE, DIMENSION(:) :: d_start, work_ns
    DOUBLE PRECISION, ALLOCATABLE, DIMENSION(:) :: start, finish
    DOUBLE PRECISION, ALLOCATABLE, DIMENSION(:,:) :: eoa

    TYPE (isat_raw_data), POINTER :: data_ptr

    LOGICAL :: within_eoa, within_finish_atol

    ALLOCATE (d_start(ns),work_ns(ns),start(ns),finish(nf))
    ALLOCATE (eoa(ns,ns))

    DO i = 1,ns
       DO j = 1,ns
          eoa(i,j) = leaf%eoa(i,j)
       ENDDO
    ENDDO

    data_ptr => isat_raw_head

    ! Run through all of the data sequentially
    DO WHILE (ASSOCIATED(data_ptr%next))

       DO i = 1,ns
          start(i) = data_ptr%start(i)
       ENDDO

       ! determine if start is within the eoa
       ! eps_sq = (start - stored_start)' * eoa * (start - stored_start)
       DO i = 1,ns
          d_start(i) = start(i) - leaf%start(i)
       ENDDO
       ! work_ns = eoa_big * d_start
       CALL DGEMV ('N', ns, ns, 1.0D0, eoa, ns, d_start, 1, &
            0.0D0, work_ns, 1)
       ! eps_sq = d_start' * work_ns
       eps_sq = 0.0D0
       DO i = 1,ns
          eps_sq = eps_sq + (d_start(i) * work_ns(i))
       ENDDO
       IF (eps_sq <= atol_sq) THEN
          within_eoa = .TRUE.
       ELSE
          within_eoa = .FALSE.
       ENDIF

       IF (within_eoa) THEN
          ! determine if finish is within the required tolerance
          CALL isat_retrieve (ns,start,nf,finish,leaf) 
       
          within_finish_atol = .TRUE.
          DO i = 1,nf
             IF (ABS(finish(i) - data_ptr%finish(i)) > atol(i)) THEN
                within_finish_atol = .FALSE.
             ENDIF
          ENDDO
 
          IF (within_finish_atol.EQV..FALSE.) THEN
             ! contract eoa
             CALL isat_growth (ns,start,atol_sq,leaf)
          ENDIF
       ENDIF

       data_ptr => data_ptr%next
    ENDDO

    DEALLOCATE (d_start,work_ns,start,finish,eoa)

  END SUBROUTINE isat_growth_check

END MODULE isat_module
