PROGRAM main

  USE files
  USE isat_module

  IMPLICIT NONE
  
  ! files
  CHARACTER (len=20) :: training_input = 'training_inputs.csv'
  CHARACTER (len=20) :: training_output = 'training_outputs.csv'
  CHARACTER (len=20) :: input = 'test_inputs.csv'
  CHARACTER (len=20) :: output = 'test_outputs.csv'

  ! pointer to the raw data storage
  TYPE (isat_raw_data), POINTER :: isat_raw_ptr

  !  TYPE (isat_btree), POINTER :: btree_ptr
  !  TYPE (isat_record), POINTER :: leaf_ptr

  !  TYPE :: isat_record
  ! isat records
  ! accessed    = number of times the record is accessed
  ! start       = vector of independent variables
  !               "start" determines "finish"
  ! finish      = vector of dependent variables
  ! sensitivity = sensitivity matrix A(i,j) = d(finish)_(i) / d(start)_(j)
  ! eoa         = ellipsoid of accuracy around "start"
  ! stem        = pointer back to the binary tree stem
  !     INTEGER :: accessed
  !     DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: start
  !     DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: finish
  !     DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: sensitivity
  !     DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: eoa
  !     TYPE (isat_btree), POINTER :: stem
  !  END TYPE isat_record
  
  !  TYPE :: isat_btree
  ! binary tree
  ! the cutting plane is defined by  v' * y0 = a
  ! v      = transpose of the vector defining the cutting plane
  ! a      = scalar as part of the definition of the cutting plane
  ! lt     = pointer to the next node if v' * y0 < a
  ! gt     = pointer to the next node if v' * y0 >= a
  ! leaf   = pointer to the index of leaves in "isat_records"
  !          this pointer is nullified if it points to another branch
  !     DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: v
  !     DOUBLE PRECISION :: a
  !     TYPE (isat_btree), POINTER :: lt
  !     TYPE (isat_btree), POINTER :: gt
  !     TYPE (isat_record), POINTER :: leaf
  !  END TYPE isat_btree

  INTEGER :: i,j,iu,opt
  INTEGER :: ns, nf
  DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: start, finish, atol
  DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: finish_real, error
  DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: a, a_real_1, a_real_2
  DOUBLE PRECISION :: sum_error, avg_error, max_error

  ! lines of training and production data
  INTEGER :: trows,prows

  ! training data
  DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: tin,tout

  ! data
  DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: pin,pout
 
  INTRINSIC RANDOM_NUMBER, RESHAPE, SUM, SIN, COS

  max_error = 0.0d0
  sum_error = 0.0d0
  
  PRINT *, 'default options set'
  isat_print_level = 5
  isat_binary_trees = 1
  isat_scalc = 1
  isat_do_growth_check = .TRUE.

  ! read in raw data

  ! training input
  PRINT *, 'read in training inputs'
  IF (.NOT.exists(training_input)) THEN
     PRINT *, TRIM(training_input), ' does not exist'
     PRINT *, 'stopping...'
     STOP
  ENDIF
  ! capture number of inputs
  ns = file_columns(training_input,',')
  ! get number of training rows
  trows = file_lines(training_input)

  ! training output
  PRINT *, 'read in training outputs'
  IF (.NOT.exists(training_output)) THEN
     PRINT *, TRIM(training_output), ' does not exist'
     PRINT *, 'stopping...'
     STOP
  ENDIF
  ! capture number of outputs
  nf = file_columns(training_output,',')
  ! check for consistency in number of training rows
  i = file_lines(training_output)
  IF (trows - i /= 0) THEN
     PRINT *, 'Inconsistent number of rows in training data'
     PRINT *, 'Training Inputs:',trows
     PRINT *, 'Training Outputs:',i
     PRINT *, 'stopping...'
     STOP
  ENDIF

  ! production input
  PRINT *, 'read test inputs'
  IF (.NOT.exists(input)) THEN
     PRINT *, TRIM(input), ' does not exist'
     PRINT *, 'stopping...'
     STOP
  ENDIF
  ! capture number of inputs
  i = file_columns(input,',')
  IF (ns - i /= 0) THEN
     PRINT *, 'Inconsistent number of columns between input training and test data'
     PRINT *, 'Training Columns:',ns
     PRINT *, 'Test Columns:',i
     PRINT *, 'stopping...'
     STOP
  ENDIF
  ! get number of training rows
  prows = file_lines(input)

  ! training output
  PRINT *, 'read in test outputs'
  IF (.NOT.exists(output)) THEN
     PRINT *, TRIM(output), ' does not exist'
     PRINT *, 'stopping...'
     STOP
  ENDIF
  ! capture number of inputs
  i = file_columns(output,',')
  IF (nf - i /= 0) THEN
     PRINT *, 'Inconsistent number of columns between output training and test data'
     PRINT *, 'Training Columns:',nf
     PRINT *, 'Test Columns:',i
     PRINT *, 'stopping...'
     STOP
  ENDIF
  ! check for consistency in number of test rows
  i = file_lines(output)
  IF (prows - i /= 0) THEN
     PRINT *, 'Inconsistent number of rows in test data'
     PRINT *, 'Test Inputs:',prows
     PRINT *, 'Test Outputs:',i
     PRINT *, 'stopping...'
     STOP
  ENDIF

  IF (isat_print_level>=1) THEN
     PRINT *, ''
     PRINT *, 'Dimensions_______________'
     PRINT *, 'Inputs       :',ns
     PRINT *, 'Outputs      :',nf
     PRINT *, 'Training Data:',trows
     PRINT *, 'Test Data    :',prows
  ENDIF

  ! allocate data arrays
  ALLOCATE (start(ns), finish(nf), atol(nf))
  ALLOCATE (finish_real(nf), error(nf))
  ALLOCATE (a(nf,ns), a_real_1(nf,ns), a_real_2(nf,ns))
  ALLOCATE (tin(trows,ns),tout(trows,nf))
  ALLOCATE (pin(prows,ns),pout(prows,nf))

  ! read data into arrays
  CALL read_csv(training_input,trows,ns,tin)
  CALL read_csv(training_output,trows,nf,tout)
  CALL read_csv(input,prows,ns,pin)
  CALL read_csv(output,prows,nf,pout)

  ! isat error tolerance (only for the "finish" values)
  atol(1:nf) = 1.d-1

  PRINT *, ''
  PRINT *, 'training...'
  DO i = 1,trows
     start(1:ns) = tin(i,1:ns)
     finish(1:nf) = tout(i,1:nf)
     opt = 6
     ! call isat, initializing it the first time through
     IF (i == 1) opt = 5
     CALL isat(ns,start,nf,finish,nf,a,atol,opt)
  ENDDO

  PRINT *, ''
  PRINT *, '___________________________________________'
  PRINT *, '*** Retrievals only - isat_results.txt'
  PRINT *, '___________________________________________'
  CALL open_file_for_write('isat_results.txt',iu)
  sum_error = 0.d0
  max_error = 0.d0
  DO i = 1,prows
     opt = 4
     start(1:ns) = pin(i,1:ns)
     finish(1:nf) = pout(i,1:nf)
     finish_real(1:nf) = finish(1:nf)
     CALL isat(ns,start,nf,finish,nf,a,atol,opt)
     WRITE(iu,*) finish(1:nf)

     ! error
     DO j = 1,nf
        error(j) = ABS(finish_real(j) - finish(j))
     ENDDO

     IF (error(1).gt.atol(1)) then
        !PRINT *, 'high approximation error: ', i, ' - ', opt, error(1)
     endif
     IF (abs(error(1)) >= max_error) max_error = abs(error(1))
     sum_error = sum_error + abs(error(1))

  ENDDO
  CLOSE(iu)

  PRINT *, ''
  PRINT *, '___________________________________________'
  PRINT *, '*** Automatic Error Control - isat_results_ec.txt'
  PRINT *, '___________________________________________'
  CALL open_file_for_write('isat_results_ec.txt',iu)
  print *, 'automatic error control - isat_results_ec.txt'
  opt = 0
  sum_error = 0.d0
  max_error = 0.d0
  DO i = 1,prows
     start(1:ns) = pin(i,1:ns)
     finish(1:nf) = pout(i,1:nf)
     finish_real(1:nf) = finish(1:nf)

     CALL isat(ns,start,nf,finish,nf,a,atol,opt)

     WRITE(iu,*) finish(1:nf)

     ! error
     DO j = 1,nf
        error(j) = ABS(finish_real(j) - finish(j))
     ENDDO

     IF (error(1).gt.atol(1)) then
        !PRINT *, 'high approximation error: ', i, ' - ', opt, error(1)
     endif
     IF (abs(error(1)) >= max_error) max_error = abs(error(1))
     sum_error = sum_error + abs(error(1))
  ENDDO

  avg_error = sum_error / 961.0D0
  print *, 'Max error:  ', max_error
  print *, 'Avg error:  ', avg_error
  print *, 'ISAT records: ', isat_records

  CALL isat_unload

  CALL EXIT(1)

END PROGRAM main
